from time import sleep

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand

from bronski.runner import JobRunner, ProgramKilled


class Command(BaseCommand):
    help = "Start a bronski beat server"

    def handle(self, *args, **options):
        model = apps.get_model(settings.CRONTAB_MODEL)
        job = JobRunner(model)
        job.setup_signals()
        self.stdout.write("Starting Bronski server...")
        job.start()

        while True:
            try:
                sleep(10)
            except ProgramKilled:
                self.stdout.write("Stopping Bronski server...")
                job.stop()
                break



from datetime import datetime, timedelta

from croniter import croniter

from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.db.models.functions import Now
from django.utils import timezone
from django.utils.module_loading import import_string


def cron_validator(value):
    if not croniter.is_valid(value):
        raise ValidationError("Invalid crontab format")


class CrontabBaseQuerySet(models.QuerySet):
    def enabled(self):
        return self.filter(is_enabled=True)

    @staticmethod
    def now():
        """
        We have fiddly requirements for `now`.

        `croniter` has issues if you pass it the current time in UTC, and will
        give you a datetime with UTC timezone, but adjusted to system local
        timezone.

        To correct for this, we set the current timezone.

        Additionally, asking for `get_next`, if the current time matches the
        crontab pattern, `croniter` will return the _next_ time.

        To correct for this, we subtract one second from the time.

        Setting microsecond to 0 is just tidiness.
        """
        return timezone.localtime(
            timezone.now().replace(microsecond=0) - timedelta(seconds=1)
        )

    def scan_jobs(self):
        """
        Convenience function for scanning a set of Jobs exclusively.

        Affects DB level locking.
        """
        with transaction.atomic():
            for job in self.select_for_update(skip_locked=True):
                yield job

    def current_jobs(self):
        """
        Yield jobs which should be run this minute.
        """
        now = self.now()

        for job in (
            self.enabled()
            .filter(last_run__lt=Now() - timedelta(seconds=59))
            .scan_jobs()
        ):
            next_run = croniter(job.crontab, now).get_next(datetime)
            if (next_run - now) < timedelta(minutes=1):
                yield job

    def missed_jobs(self):
        """
        yields jobs whose last run is longer ago than their crontab frequency.
        """
        now = self.now()

        for job in self.enabled().scan_jobs():
            next_run = croniter(job.crontab, job.last_run).get_next(datetime)
            if next_run < now:
                yield job


class CrontabBase(models.Model):
    """
    Abstract base model for custom Bronski crontab model.
    """

    crontab = models.CharField(max_length=100, validators=[cron_validator])
    function = models.CharField(max_length=255)
    kwargs = models.JSONField(default=dict, blank=True)
    is_enabled = models.BooleanField(default=True)
    last_run = models.DateTimeField(default=timezone.now)

    objects = CrontabBaseQuerySet.as_manager()

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.function} @ {self.crontab}"

    def get_function(self):
        """Helper to import the named function in `function`."""
        return import_string(self.function)

    def get_kwargs(self):
        """Hepler to return the kwargs field or default to empty dict."""
        return self.kwargs or {}

    def run(self):
        func = self.get_function()
        kwargs = self.get_kwargs()
        func(**kwargs)

    def mark_run(self):
        """
        Update the last_run to Now, and save.
        """
        self.last_run = Now()
        self.save()
        self.refresh_from_db()



from django.contrib import admin
from django.db.models.functions import Now


def trigger_job(self, request, queryset):
    for job in queryset:
        job.run()

    queryset.update(last_run=Now())


trigger_job.short_description = "Run Selected..."


class CrontabBaseAdmin(admin.ModelAdmin):
    list_display = (
        "crontab",
        "function",
        "kwargs",
        "is_enabled",
        "last_run",
    )
    list_display_links = ("function",)
    list_filter = ("is_enabled",)
    list_editable = ("is_enabled",)
    date_hierarchy = "last_run"

    actions = [
        trigger_job,
    ]



import logging
import signal
import threading
from datetime import datetime

import django
from django.db import close_old_connections
from django.db.models.functions import Now

django.setup()

log = logging.getLogger(__name__)


class ProgramKilled(Exception):
    pass


class JobRunner(threading.Thread):
    def __init__(self, model):
        super().__init__()
        self.trigger = threading.Event()
        self.stopped = threading.Event()

        self.model = model

    def setup_signals(self):
        # Set up escape hatch
        signal.signal(signal.SIGTERM, self.break_handler)
        signal.signal(signal.SIGINT, self.break_handler)

        # Set up handler for signal
        signal.signal(signal.SIGALRM, self.timer_handler)

        # Align to the minute boundary.
        offset = 60 - datetime.now().second
        # Set OS timer to bug us every minute
        signal.setitimer(signal.ITIMER_REAL, offset, 60)

    def timer_handler(self, signum, frame):
        """Signal handler for timer."""
        log.debug("Scanning jobs...")
        self.trigger.set()

    @staticmethod
    def break_handler(signum, frame):
        """Signal handler for ctrl-c / SIGTERM, to exit program."""
        log.info("Detected stop request...")
        raise ProgramKilled

    def stop(self):
        log.info("Setting stop...")
        self.stopped.set()
        self.trigger.set()
        self.join()

    def run(self):
        log.info("Starting Bronski Runner...")
        while self.trigger.wait():
            if self.stopped.is_set():
                log.info("Stopping...")
                break

            for job in self.model.objects.current_jobs():
                log.info("Running job: %s", job)
                try:
                    job.run()
                    job.last_run = Now()
                    job.mark_run()
                except Exception:
                    log.exception("Failed running job: %s", job)

            # Clean up django DB connections
            close_old_connections()

            # Reset our trigger
            self.trigger.clear()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bronski', 'bronski.management', 'bronski.management.commands']

package_data = \
{'': ['*']}

install_requires = \
['croniter>=0.3.31,<0.4.0', 'django>=3.1']

setup_kwargs = {
    'name': 'bronski',
    'version': '0.4.0',
    'description': 'Simple, generic cron-like job runner service for Django',
    'long_description': '=======\nBronski\n=======\n\n.. image:: https://badge.fury.io/py/bronski.svg\n    :target: https://pypi.org/project/bronski\n    \n.. image:: https://img.shields.io/pypi/pyversions/bronski.svg\n    :target: https://pypi.org/project/bronski\n    \n.. image:: https://github.com/uptick/bronski/workflows/Test/badge.svg\n    :target: https://github.com/uptick/bronski/actions?query=workflow%3ATest\n\n.. rubric:: A beat server for Django, with cron-like syntax\n\nBronski allows you to configure periodic function calls using a Django model.\n\nIt is ideally suited to being a task "beat" sever, akin to celery-beat.\n\nInstall\n-------\n\n.. code-block:: sh\n\n    $ pip install bronski\n\n\nSetup\n-----\n\n1. Add \'bronski\' to your ``INSTALLED_APPS``\n\n   This is only needed to enable the management command.\n\n2. Create a model in your own app that inherits from ``bronski.models.CrontabBase``\n\n3. Create and apply migrations:\n\n   .. code-block:: sh\n\n    $ manage.py makemigrations\n    $ manage.py migrate\n\n4. Specify your model in settings\n\n   .. code-block:: python\n\n    CRONTAB_MODEL = "myapp.MyCronModel"\n\n5. Launch your beat server:\n\n   .. code-block:: sh\n\n    $ ./manage.py bronski\n\nEach minute the ``bronski`` service will scan the model for enabled jobs that\nhaven\'t been run in the past 59 seconds. It will then check each to see if its\ncrontab definition matches the next minute.\n\nFor job records that match, their ``run`` method will be called. The default\n``run`` method will:\n\n- get the specified function by calling ``self.get_function()``\n- resolve the ``kwargs`` to use by calling ``self.get_kwargs()``\n- invoke the function with the ``kwargs``.\n\nYou can override ``run`` in your custom model to, for instance, enqueue jobs:\n\n.. code-block:: python\n\n    class Jobs(CrontabBase):\n\n        def run(self):\n            func = self.get_function()\n            kwargs = self.get_kwargs()\n\n            # Celery task API:\n            func.delay(**kwargs)\n            # Dramatiq actor API:\n            func.send(**kwargs)\n',
    'author': 'Curtis Maloney',
    'author_email': 'curtis@tinbrain.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/uptick/bronski',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
